;;; ihtml (markup helping tool for iso-html or xhtml11)
;;; 2006-05-19, 2021-09-08 Ver.0.2.6.20210908
;;; http://purl.com/net/dkcc/shed/ihtml.html
;;; License: CC0



;;; (6) package
;;; 2006-05-21, 2006-05-23

(provide "ihtml")
#|
(in-package "editor")
(export
 '())
|#



;;; (1) ihtml-relative-path
;;; 2006-05-19

(defun ihtml-relative-path (href &optional (base (directory-namestring (default-directory)))) ; href must be available, absolute path
  "Return relative path. href must be available, absolute path. e.g. (ihtml-relative-path \"D:/diary/index.html#contents\")"
  (if (path-equal href base) "./"
    (let ((dirp (file-directory-p href))
          (href (split-string href "/"))
          (base (split-string base "/")))
      (while (string= (car href) (car base))
        (pop href)
        (pop base))
      (fill base "..")
      (setq base (append base href))
      (format nil "~A~{/~A~}~A" (pop base) base (if dirp "/" "")))))

#|

;;; test
(let ((href '("I:" "My Documents" "Diary" "shed" "Kaeisuisyou.html#FILES"))
      (base '("I:" "My Documents" "Diary" "kotoba")))
  (while (string= (car href) (car base))
    (pop href)
    (pop base))
  (fill base "..")
  (nconc base href)
  (format nil "~A~{/~A~}" (pop base) base))

(let ((href '("I:" "My Documents" "Diary" "shed" "Kaeisuisyou.html#FILES"))
      (base '("I:" "My Documents" "Diary" "shed")))
  (while (string= (car href) (car base))
    (pop href)
    (pop base))
  (fill base "..")
  (setq base (append base href)))

(split-string "I:/My Documents/Diary/shed/Kaeisuisyou.html#FILES" "/")
("I:" "My Documents" "Diary" "shed" "Kaeisuisyou.html#FILES")
(split-string "I:/My Documents/Diary/kotoba" "/")
("I:" "My Documents" "Diary" "kotoba")

(ihtml-relative-path "I:/My Documents/Diary/shed/Kaeisuisyou.html#FILES" "I:/My Documents/Diary/kotoba")
 => "../shed/Kaeisuisyou.html#FILES"
(ihtml-relative-path "I:/My Documents/Diary/shed/Kaeisuisyou.html" "I:/My Documents/Diary/kotoba")
 => "../shed/Kaeisuisyou.html"
(ihtml-relative-path "I:/My Documents/Diary" "I:/My Documents/Diary/")
 => "../Diary" ; Not problem, because of file name collision on Windows

(values
 (ihtml-relative-path "Z:/LHD-LAN.TXT")
 (ihtml-relative-path "Z:/TEMP_WXP")
 (ihtml-relative-path "Z:/_CACHE/20060323/CLOS.L")
 (ihtml-relative-path "Z:/_CACHE/20060323/SCORE/")
 (ihtml-relative-path "Z:/_CACHE/20060323"))
"../../LHD-LAN.TXT"
"../../TEMP_WXP/"
"CLOS.L"
"SCORE/"
"./"
|#



;;; (2) ihtml-markup
;;; 2006-05-20, 2011-10-13

(defun ihtml-markup (start end tag &optional attr)
  "Markup between start and end. You can give attributes by plist attr."
  (when (> start end) (rotatef start end))
  (goto-char end)
  (insert "</" tag ">")
  (goto-char start)
  (insert "<" tag (format nil "~{ ~A~@[=\"~A\"~]~}" (mapcar #'ihtml-auto-escape attr)) ">"))
(defun ihtml-markup-selection (tag &optional attr)
  "Markup selection. e.g. (ihtml-markup-selection \"a\" '(\"href\" \"http://purl.com/dkcc/\" \"title\" \"DKCC Plaza\"))"
  (interactive "*stag: ")
  (if (get-selection-type)
      (ed::map-selection
       (lambda (start end)
         (ihtml-markup start end tag attr)))
    (ihtml-markup (point) (point) tag attr)))
(defun ihtml-markup-empty (tag &optional attr (xhtml *ihtml-xhtml11*))
  "Insert empty element."
  (let ((nestc (if xhtml " /" "")))
    (goto-char (point))
    (insert "<" tag (format nil "~{ ~A~@[=\"~A\"~]~}" (mapcar #'ihtml-auto-escape attr)) nestc ">")))

#|
(defun ihtml-markup-selection (tag &rest attr)
;  (interactive "*stag: \nsattr: ")
  (ed::map-selection
   #'(lambda (start end)
       (when (> start end) (rotatef start end))
       (save-excursion
         (goto-char end)
         (insert "</" tag ">")
         (goto-char start)
         (insert "<" tag (format nil "~{ ~A=\"~A\"~}" attr) ">")))))
|#

#|
12[345678]90
+ (ihtml-markup-selection "em" "id=\"N50\" class=\"example\"")
or (ihtml-markup-selection "em" "id" "N50" "class" "example")
or (ihtml-markup-selection "em" '("id" "N50" "class" "example"))
 => 12<em id="N50" class="example">345678</em>90

(ihtml-markup-selection "em" (list "id" "N50"))
<em id="N50">(ihtml-markup-selection "em" (list "id" "N50"))</em>
|#



;;; (3) ihtml-find-file-hook
;;; 2006-05-20, 2011-10-08

;(defvar-local *ihtml-xhtml11* nil)
(defvar *ihtml-xhtml11* nil
  "Determine conforming specification. If t, XHTML 1.1. If nil, ISO-HTML.")
(defvar *ihtml-img-extensions* (compile-regexp "\\.\\(bmp\\|gif\\|jpeg?\\|jpg\\|png\\|svgz?\\|tiff?\\)$")
  "Extensions (regexp) of image file.")
#| ;;; => ihtml-string-to-uri
(defun ihtml-file-to-href (file)
  (si::www-url-encode (ihtml-relative-path (map-backslash-to-slash file)) nil "0-9A-Za-z---._~/"))
|#
(defun ihtml-find-file-hook (file)
  "If selection exists, Markup <a>. If not, insert <img> when filename matches *ihtml-img-extensions*."
;  (when (string-matchp "\\.\\(html?\\|xht\\|xhtml?\\)" (string (get-buffer-file-name)))
  (when (member buffer-mode '(html-mode html+-mode))
    (let ((href (ihtml-string-to-uri file)))
      (cond
       ((get-selection-type)
        (ihtml-markup-selection "a" (list "href" href))
        (selected-buffer))
       ((string-matchp *ihtml-img-extensions* href)
        (ihtml-markup-empty "img" (list "alt" (read-string "alt: ") "src" href) *ihtml-xhtml11*)
        (selected-buffer))))))
(add-hook '*before-find-file-hook* 'ihtml-find-file-hook)
(defun ihtml-toggle-xhtml11 ()
  "Toggle *ihtml-xhtml11*."
  (interactive)
  (ed::toggle-mode '*ihtml-xhtml11* nil nil)
  (message "*ihtml-xhtml11*: ~S" *ihtml-xhtml11*)
  *ihtml-xhtml11*)



;;; (4) inline elements
;;; 2006-05-21, 2014-10-28

(defmacro ihtml-defun-markup-selection (tag &optional attr &key doc)
  "Define markup function of inline element. e.g. (ihtml-defun-markup-selection abbr (list \"title\" (read-string \"title: \")))"
  (let ((tag (string tag)))
    `(defun ,(intern (concat "ihtml-" tag)) ()
       (interactive "*")
       ,doc
       (ihtml-markup-selection ,tag ,attr))))
(defmacro ihtml-defun-markup-empty (tag &optional attr &key block doc)
  "Define markup function of empty element. e.g. (ihtml-defun-markup-empty hr nil :block t)"
  (let ((tag (string tag)))
    `(defun ,(intern (concat "ihtml-" tag)) ()
       (interactive "*")
       ,doc
       (ihtml-markup-empty ,tag ,attr)
       ,(when block '(newline-and-indent)))))

#|
(macroexpand-1 '(ihtml-defun-markup-selection "em"))
(defun ihtml-em nil
  (interactive "*")
  (ihtml-markup-selection "em" nil))

(macroexpand-1 '(ihtml-defun-markup-selection "ins" (list "datetime" (format-date-string "%Y-%m-%dT%H:%M:%S%#Z"))))
(defun ihtml-ins nil
  (interactive "*")
  (ihtml-markup-selection "ins" (list "datetime" (format-date-string "%Y-%m-%dT%H:%M:%S%#Z"))))
|#

(dolist (tag '(cite code dfn em kbd samp strong var b i sub sup tt))
  (eval `(ihtml-defun-markup-selection ,tag nil :doc ,(format nil "Markup selection by <~A>." tag))))
(dolist (tag '(abbr acronym))
  (eval `(ihtml-defun-markup-selection ,tag (list "title" (read-string "title: ")) :doc ,(format nil "Markup selection by <~A title=\"...\">." tag))))
(dolist (tag '(ins del))
  (eval `(ihtml-defun-markup-selection ,tag (list "datetime" (format-date-string "%Y-%m-%dT%H:%M:%S%#Z")) :doc ,(format nil "Markup selection by <~A>." tag))))
;;;(ihtml-defun-markup-selection a (list "href" (ihtml-string-to-uri (read-string "href: " :default ""))))
;;;(ihtml-defun-markup-selection a (list "href" (ihtml-string-to-uri (read-file-name "href: " :default ""))))
(defun ihtml-a-global ()
  (interactive "*")
  "Markup selection by <a>."
  (ihtml-markup-selection "a" (list "href" (ihtml-string-to-uri (read-string "href: " :default "")))))
(defun ihtml-a-local ()
  (interactive "*")
  "Markup selection by <a>, for local address (relative path)."
  (ihtml-markup-selection "a" (list "href" (ihtml-string-to-uri (read-file-name "href: " :default "")))))
(defvar *ihtml-a-local* nil
  "If t, ihtml-a calls ihtml-a-local. If nil, ihtml-a calls ihtml-a-global.")
(defun ihtml-a ()
  "Markup selection by <a>."
  (interactive "*")
  (if *ihtml-a-local* (ihtml-a-global) (ihtml-a-local)))
(ihtml-defun-markup-empty img (list "alt" (read-string "alt: ") "src" (ihtml-string-to-uri (read-file-name "src: "))) :doc "Insert <img alt=\"...\" src=\"...\">.")
(ihtml-defun-markup-selection object (list "data" (ihtml-string-to-uri (read-file-name "data: "))) :doc "Insert <object data=\"...\"></object>.")
(ihtml-defun-markup-empty param (list "name" (read-string "name: ") "value" (read-string "value: ")) :doc "Insert <param name=\"...\" value=\"...\">.")
(ihtml-defun-markup-selection q
  (let ((cite (read-string "cite: ")) (title (read-string "title: ")) attr)
    (when (string/= title "") (push title attr) (push "title" attr))
    (when (string/= cite "") (push cite attr) (push "cite" attr))
    attr)
  :doc "Markup selection by <q cite=\"...\" title=\"...\">.")
(ihtml-defun-markup-selection span
  (let ((class (read-string "class: ")) (title (read-string "title: ")) attr)
    (when (string/= title "") (push title attr) (push "title" attr))
    (when (string/= class "") (push class attr) (push "class" attr))
    attr)
  :doc "Markup selection by <span title=\"...\" class=\"...\">.")
(ihtml-defun-markup-selection bdo (list "dir" (completing-read "dir: " '(ltr rtl) :default "ltr" :must-match t)) :doc "Markup selection by <bdo dir=\"ltr\"> or <bdo dir=\"rtl\">.")

#|
(dolist (x '("c:/winnt/" "c:/winnt/nul" "c:/winntt/" "c:/winntt/nul" "http://example.com/.htaccess" "mailto:admin@example.com"))
  (format t "\n~A\n" x)
   (dolist (f '(check-valid-pathname file-exist-p valid-path-p))
     (format t "\t~A: ~S\n" f (handler-case (funcall f x) (error (c) c)))))


c:/winnt/
	check-valid-pathname: t
	file-exist-p: t
	valid-path-p: t

c:/winnt/nul
	check-valid-pathname: t
	file-exist-p: t
	valid-path-p: t

c:/winntt/
	check-valid-pathname: nil
	file-exist-p: nil
	valid-path-p: t

c:/winntt/nul
	check-valid-pathname: #S(path-not-found pathname "c:/winntt/nul" datum w肳ꂽpX܂B )
	file-exist-p: nil
	valid-path-p: nil

http://example.com/.htaccess
	check-valid-pathname: #S(path-not-found pathname "http://example.com/.htaccess" datum w肳ꂽpX܂B )
	file-exist-p: nil
	valid-path-p: nil

ftp://example.com
	check-valid-pathname: nil
	file-exist-p: nil
	valid-path-p: t

mailto:admin@example.com
	check-valid-pathname: #S(file-error pathname "mailto:admin@example.com" datum t@CAfBNgA܂̓{[ x̍\ԈĂ܂B )
	file-exist-p: nil
	valid-path-p: nil
nil

|#

(dolist (tag '(br hr))
  (eval `(ihtml-defun-markup-empty ,tag () :block t :doc ,(format nil "Insert <~A>." tag))))

#|
; special var.
(defvar *ihtml-special* '(a bdo br img object map q span))
(defvar *ihtml-logical-styles* '(abbr acronym cite code dfn em kbd samp strong var))
(defvar *ihtml-physical-styles* '(b i sub sup tt))
(defvar *ihtml-block* '(blockquote div dl fieldset form hr ol p pre table ul))
(defvar *ihtml-form-fields* '(button input label select textarea))

(defvar *ihtml-core* '(class id title))
(defvar *ihtml-i18n* '(dir lang))
(defvar *ihtml-shape* '(circle default poly rect))
(defvar *ihtml-InputType* '(checkbox file hidden password radio reset submit text))


; use hash table?
(defvar *ihtml-entity-table* nil)
(unless *ihtml-entity-table*
  (setq *ihtml-entity-table* (make-hash-table))
  (setf (gethash 'logical.styles *ihtml-entity-table*) '(abbr acronym cite code dfn em kbd samp strong var))
  (setf (gethash 'physical.styles *ihtml-entity-table*) '(b i sub sup tt)))

; test
(hash-table-all-entry *ihtml-entity-table*)
1. logical.styles
	(abbr acronym cite code dfn em kbd samp strong var)
2. physical.styles
	(b i sub sup tt)
|#



;;; (5) *ihtml-popup-menu*
;;; 2006-05-21, 2011-11-01
;;; error occur in siteinit.l, so for .xyzzy only
;;; -> moved to ihtml-menu.l



;;; (7) ihtml-current-heading
;;; 2006-05-23

(defun ihtml-current-heading ()
  "Return current heading level (nil, 1-6)."
  (save-excursion
    (while (scan-buffer (compile-regexp "</h\\([1-6]\\)>" t) :reverse t :no-dup t)
      (unless (eq (parse-point-syntax) :comment)
        (return (parse-integer (match-string 1)))))))

#|
</h4>
<!--;;; </h5> -->

(ihtml-current-heading)
|#



;;; (8) ihtml-string-to-uri
;;; 2006-05-24, 2014-10-28

(defvar *ihtml-string-to-uri-escape* t
  "If t, si::www-url-encode is applied in ihtml-string-to-uri.")
(defun ihtml-string-to-uri (str)
  "Convert str to uri."
  (if (file-exist-p str)
      (if *ihtml-string-to-uri-escape*
          (si::www-url-encode (ihtml-relative-path (map-backslash-to-slash str)) nil "0-9A-Za-z---._~/")
        (ihtml-relative-path (map-backslash-to-slash str)))
;    (substitute-string str "\"" "&quot;")))
    str))
;;; problem: read-file-name regards "?" as wildcard.



;;; (9) ihtml-markup-line
;;; 2006-05-25, 2014-10-30

(defvar *ihtml-line-break* t
  "If t, ihtml-markup-line breaks line.")
(defun ihtml-toggle-line-break ()
  "Toggle *ihtml-line-break*."
  (interactive)
  (ed::toggle-mode '*ihtml-line-break* nil nil)
  (message "*ihtml-line-break*: ~S" *ihtml-line-break*)
  *ihtml-line-break*)
(defun ihtml-markup-line (tag &optional attr &key #|(keep-line t) |#(break-line *ihtml-line-break*) (permit-empty t))
  "Markup current line and go to next line."
  (interactive "*stag: ")
  (if (get-selection-type) (ihtml-markup-selection tag attr)
    (let (start end)
;;;      (when keep-line (goto-eol))
      (goto-eol)
      (setq end (point))
      (when (or break-line (null (next-line)))
        (newline)
        (when mode-specific-indent-command (funcall mode-specific-indent-command)))
      (save-excursion
        (previous-line)
        (goto-bol)
        (setq start (point))
        (save-restriction
          (narrow-to-region start end)
          (skip-white-forward)
          (setq start (point))
          (when (or (/= start end) permit-empty)
            (ihtml-markup start end tag attr)))))))
#|
(defun ihtml-markup-line (tag &optional attr keep-line)
  (interactive "*")
  (if (get-selection-type) (ihtml-markup-selection tag attr)
    (let (start end)
      (when keep-line (goto-eol))
      (setq end (point))
      (newline)
      (when mode-specific-indent-command (funcall mode-specific-indent-command))
      (save-excursion
        (previous-line)
        (goto-bol)
        (setq start (point))
        (save-restriction
          (narrow-to-region start end)
          (skip-white-forward)
          (setq start (point))
          (unless (= start end)
            (ihtml-markup start end tag attr)))))))
|#
(defmacro ihtml-defun-markup-line (tag &optional attr &key doc)
  "Define markup function of element for line. e.g. (ihtml-defun-markup-line div (list \"class\" (read-string \"class: \")))"
  (let ((tag (string tag)))
    `(defun ,(intern (concat "ihtml-" tag)) ()
       (interactive "*")
       ,doc
       (ihtml-markup-line ,tag ,attr))))



;;; (10) block elements
;;; 2006-05-25, 2014-10-28

(dolist (h '(1 2 3 4 5 6))
  (eval
   `(defun ,(intern (format nil "ihtml-h~D" h)) ()
      ,(format nil "Markup current line by <h~D>." h)
      (interactive "*")
      (let ((level (ihtml-current-heading)))
        (unless level (setq level 0))
        (if (> ,(1- h) level) (error "cannot skip heading level: current heading level is ~D" level)
          (ihtml-markup-line ,(format nil "h~D" h)))))))
(dolist (tag '(address p pre))
  (eval
   (let ((tag (string tag)))
     `(defun ,(intern (concat "ihtml-" tag)) ()
        ,(format nil "Markup current line by <~A>." tag)
        (interactive "*")
        (ihtml-markup-line ,tag)))))
(defun ihtml-div ()
  "Markup current line by <div class=\"...\">."
  (interactive "*")
  (let ((class (read-string "class: ")) attr)
    (when (string/= class "") (push class attr) (push "class" attr))
    (ihtml-markup-line "div" attr)))
(defun ihtml-blockquote ()
  "Markup current line by <blockquote cite=\"...\" title=\"...\">."
  (interactive "*")
  (let ((cite (read-string "cite: ")) (title (read-string "title: ")) attr)
    (when (string/= title "") (push title attr) (push "title" attr))
    (when (string/= cite "") (push cite attr) (push "cite" attr))
    (ihtml-markup-line "blockquote" attr)))
(defun ihtml-h0 ()
  "Markup current line by <hn>, where n: current heading level."
  (interactive "*")
  (let ((level (ihtml-current-heading)))
    (if level (ihtml-markup-line (format nil "h~D" level))
      (ihtml-markup-line "h1"))))
(defun ihtml-h- ()
  "Markup current line by <hn->, where n-: current heading level - 1."
  (interactive "*")
  (let ((level (ihtml-current-heading)))
    (unless level (setq level 0))
    (if (> level 1) (ihtml-markup-line (format nil "h~D" (1- level)))
      (ihtml-markup-line "h1"))))
(defun ihtml-h+ ()
  "Markup current line by <hn+>, where n+: current heading level + 1."
  (interactive "*")
  (let ((level (ihtml-current-heading)))
    (unless level (setq level 0))
    (if (< level 6) (ihtml-markup-line (format nil "h~D" (1+ level)))
      (ihtml-markup-line "h6"))))



;;; (11) ihtml-set-attribute
;;; 2006-05-26, 2014-10-29

(defun ihtml-set-attribute (attr &optional (xhtml *ihtml-xhtml11*) &key empty-value completions must-match)
  "Set attributes of current/previous tag."
  (interactive "*sattr: ")
;  (while (scan-buffer (compile-regexp (format nil "\\(<[^/].*?\\)\\(~A=\"\\(.*?\\)\"\\)?\\(.*?>\\)" attr) (not xhtml)) :reverse t)
  (save-excursion
    (forward-char)
    (while (scan-buffer (compile-regexp "<[^/\n].*?\\(/?\\)>" (not xhtml)) :reverse t :no-dup t)
      (unless (eq (parse-point-syntax) :comment)
;        (msgbox "~S" (parse-point-syntax))
        (return
;         (reverse-region (match-beginning 0) (match-end 0))
         (let ((stag (match-data)))
           (save-restriction
             (narrow-to-region (match-beginning 0) (match-end 0))
;         (if (match-string 2)
             (if (scan-buffer (compile-regexp (format nil "[ \t\n]*~A=\"\\(.*?\\)\"" attr) (not xhtml)))
                 (progn
;                   (msgbox "~A~%~A" (match-string 0) (match-string 1))
                   (setq stag (match-data))
;                   (replace-match (concat "\\1" (read-string (format nil "~A: " attr) :default (match-string 2)) "\\4")))
                   (let ((attr-val (ihtml-auto-escape
                                    (if completions
                                        (completing-read (format nil "~A: " attr) completions :default (ihtml-auto-unescape (match-string 1)) :must-match must-match)
                                      (read-string (format nil "~A: " attr) :default (ihtml-auto-unescape (match-string 1)))))))
                     (store-match-data stag)
;                     (msgbox "~A~%~A" (match-string 0) (match-string 1))
                     (replace-match
                      (if (and (string= attr-val "") (not empty-value)) ""
                        (format nil " ~A=\"~A\"" attr attr-val)))))
               (progn
                 (store-match-data stag)
                 (if (match-string 1) ; if "/>" is contained
                     (if (syntax-whitespace-p (char-before (match-beginning 1)))
                         (goto-char (1- (match-beginning 1))) ; <hr /> -> <hr| />
                       (goto-char (match-beginning 1))) ; <hr/> -> <hr|/>
                   (goto-char (1- (match-end 0)))) ; <div> -> <div|>
                 (let ((attr-val (ihtml-auto-escape (read-string (format nil "~A: " attr)))))
                   (unless (and (string= attr-val "") (not empty-value))
                     (insert (format nil " ~A=\"~A\"" attr attr-val)))))))))))))
;             (clear-reverse-region))))))))
(defun ihtml-attr-core ()
  "Set class, id, and title of current/previous tag."
  (interactive "*")
  (dolist (attr '(class id title))
    (ihtml-set-attribute (string attr) *ihtml-xhtml11*)))
(dolist (attr '(class id title lang))
  (eval
   `(defun ,(intern (format nil "ihtml-attr-~A" attr)) ()
      (interactive "*")
      (ihtml-set-attribute ,(string attr) *ihtml-xhtml11*))))
(defun ihtml-attr-bdo ()
  "Set bdo of current/previous tag."
  (interactive "*")
  (ihtml-set-attribute "bdo" *ihtml-xhtml11* :completions '("ltr" "rtl") :must-match t))

; ihtml-dir

#|
<h2>foo</h2>

<p>bar</p>

(let ((attr "title") (xhtml nil))
  (scan-buffer (compile-regexp (format nil "<[^/\n].*?>" attr) (not xhtml)) :reverse t)
  (match-string 0))
"<[^/\\n].*?>"
|#



;;; (12) keymap
;;; 2007-08-13, 2021-09-08

(defvar *ihtml-keymap* (make-sparse-keymap)
  "Keymap for ihtml.")
(defun ihtml-define-keymap (&optional (keymap *ihtml-keymap*))
  "Define keys for ihtml."
  (interactive)
  ;(let ((keymap *global-keymap*)) ; 2021-09-08 *global-keymap* -> define-key *ihtml-keymap*
  (define-key keymap #\LFD 'ihtml-p) ; C-RET
  (define-key keymap #\F13 'ihtml-br) ; S-RET
  (define-key keymap #\M-RET 'ihtml-attr-core)
  (define-key keymap #\C-t 'ihtml-completing-markup)
  (define-key keymap #\C-0 'ihtml-h0)
  (define-key keymap #\F14 'ihtml-h0) ; C-S-RET
  (define-key keymap #\C-+ 'ihtml-h+)
  (define-key keymap #\C-- 'ihtml-h-)
  (define-key keymap #\C-1 'ihtml-h1)
  (define-key keymap #\C-2 'ihtml-h2)
  (define-key keymap #\C-3 'ihtml-h3)
  (define-key keymap #\C-4 'ihtml-h4)
  (define-key keymap #\C-5 'ihtml-h5)
  (define-key keymap #\C-6 'ihtml-h6)
  (define-key keymap #\C-* 'ihtml-li)
  (define-key keymap #\M-Up 'ihtml-sup)
  (define-key keymap #\M-Down 'ihtml-sub)
  (define-key keymap #\M-1 'ihtml-comment)
  (define-key keymap #\C-! 'ihtml-comment)
  (define-key keymap #\M-2 'ihtml-quot)
  (define-key keymap #\M-6 'ihtml-&)
  (define-key keymap #\M-7 'ihtml-apos)
  (define-key keymap #\M-\, 'ihtml-<)
  (define-key keymap #\M-. 'ihtml->)
  (define-key keymap #\M-e 'ihtml-em)
  ;  (define-key keymap #\C-e 'ihtml-escape-selection)
  ;  (define-key keymap #\C-E 'ihtml-unescape-selection)
  ;  (define-key keymap #\M-r 'ihtml-ruby)
  ;  (define-key keymap #\C-q 'ihtml-blockquote)
  ;  (define-key keymap #\M-q 'ihtml-q)
  ;  (define-key keymap #\C-t 'ihtml-markup-line)
  ;  (define-key keymap #\M-t 'ihtml-markup-selection)
  (set-extended-key-translate-table exkey-S-return #\F13)
  (set-extended-key-translate-table exkey-S-C-return #\F14)
  (set-extended-key-translate-table exkey-M-return #\M-RET))



;;; (13) ihtml-line
;;; 2007-08-14

#|
(defun ihtml-line ()
  (interactive)
  (let (start (end (point)) eol)
    (save-excursion
      (goto-bol)
      (backward-char)
      (skip-white-backward)
      (forward-char)
      (setq eol (point))
      (goto-bol)
      (setq start (point))
      (msgbox "start: ~S eol: ~S end: ~S
start-eol: ~S
eol-end:   ~S" start eol end (buffer-substring start eol) (buffer-substring eol end)))))
|#



;;; (14) ihtml-comment
;;; 2007-09-09, 2014-10-28

(defun ihtml-comment ()
  "Comment out selection. Uncomment selection if selection contains comment, or current point is in comment."
  (interactive "*")
  (if (eq (parse-point-syntax) :comment)
      (save-excursion
        (when (scan-buffer (compile-regexp "<![ \t\n]*--") :reverse t)
          (delete-region (match-beginning 0) (match-end 0)))
        (when (scan-buffer (compile-regexp "--[ \t\n]*>"))
          (delete-region (match-beginning 0) (match-end 0))))
    (if (get-selection-type)
        (ed::map-selection
         (lambda (start end)
           (save-excursion
             (when (> start end) (rotatef start end))
             (save-restriction
               (narrow-to-region start end)
               (goto-char (point-min))
               (when (zerop (replace-buffer "\\(<![ \t\n]*--\\|--[ \t\n]*>\\)" "" :regexp t))
                 (goto-char end)
                 (insert "-->")
                 (goto-char start)
                 (insert "<!--"))))))
      (let ((point (point)))
        (goto-char point)
        (insert "-->")
        (goto-char point)
        (insert "<!--")))))
(dolist (raw-esc '(("<" . "&lt;") (">" . "&gt;") ("&" . "&amp;") ("apos" . "&apos;")))
  (eval
   (let ((raw (car raw-esc))
         (esc (cdr raw-esc)))
     `(defun ,(intern (concat "ihtml-" raw)) ()
        ,(format nil "Insert ~A." esc)
        (interactive "*")
        (insert ,esc)))))
(defun ihtml-quot ()
  "Insert &quot;."
  (interactive "*")
  (insert "&quot;"))



;;; (15) list elements
;;; 2007-09-11, 2014-10-28

(dolist (tag '(li dt dd))
  (eval
   (let ((tag (string tag)))
     `(defun ,(intern (concat "ihtml-" tag)) ()
        ,(format nil "Markup current line by <~A>." tag)
        (interactive "*")
        (ihtml-markup-line ,tag)))))
(dolist (tag '(ol ul dl))
  (eval `(ihtml-defun-markup-selection ,tag nil :doc ,(format nil "Markup selection by <~A>." tag))))



;;; (16) ihtml-escape
;;; 2007-09-11, 2014-10-28

(defvar *ihtml-escape-attr-value* t
  "If t, attr values are automatically escaped.")
(defun ihtml-escape (str)
  "Escape str. e.g. \"&\" -> \"&amp;\""
  (when str
    (substitute-string
     (substitute-string
      (substitute-string
       (substitute-string
        str
        "&" "&amp;")
       "<" "&lt;")
      ">" "&gt;")
     "\"" "&quot;")))
(defun ihtml-unescape (str)
  "Unescape str. e.g. \"&amp;\" -> \"&\""
  (when str
    (substitute-string
     (substitute-string
      (substitute-string
       (substitute-string
        str
        "&amp;" "&")
       "&lt;" "<")
      "&gt;" ">")
     "&quot;" "\"")))
#|
(defmacro ihtml-auto-escape (str) ; side effect: this macro abandons match-data => use store-match-data
  `(if ,*ihtml-escape-attr-value* (ihtml-escape ,str) ,str))
(defmacro ihtml-auto-unescape (str)
  `(if ,*ihtml-escape-attr-value* (ihtml-unescape ,str) ,str))
|#
(defun ihtml-auto-escape (str) ; side effect: this function abandons match-data => use store-match-data
  "Escape attr when *ihtml-escape-attr-value* is t."
  (if *ihtml-escape-attr-value* (ihtml-escape str) str))
(defun ihtml-auto-unescape (str)
  "Unescape attr when *ihtml-escape-attr-value* is t."
  (if *ihtml-escape-attr-value* (ihtml-unescape str) str))



;;; (17) ihtml-completing-markup
;;; 2007-09-12, 2011-10-13

;;; http://purl.org/NET/ISO+IEC.15445/15445.html#DTD
(defvar *ihtml-tags* '(html head title link meta style body h1 h2 h3 h4 h5 h6 div address p ol ul li dl dt dd pre blockquote q form fieldset input button label legend select optgroup option textarea hr table caption thead tfoot tbody col colgroup tr th td abbr acronym cite code dfn em kbd samp strong var b i sub sup tt a img object param br map area span del ins bdo)
  "All tags in ISO-HTML.")
(defvar *ihtml-tags-empty* '(link meta input hr col img param br area)
  "Empty emelent tags in ISO-HTML.")
(defvar *ihtml-tags-xhtml11* '(body head html title abbr acronym address blockquote br cite code dfn div em h1 h2 h3 h4 h5 h6 kbd p pre q samp span strong var a dl dt dd ol ul li object param b big hr i small sub sup tt del ins bdo button fieldset form input label legend select optgroup option textarea caption col colgroup table tbody td tfoot th thead tr img area map meta noscript script style link base ruby rbc rtc rb rt rp)
  "All tags in XHTML 1.1.")     ; Source: The XHTML 1.1 Document Type http://www.w3.org/TR/xhtml11/doctype.html
(defvar *ihtml-tags-empty-xhtml11* '(br param hr input col img area meta link base)
  "Empty emelent tags in XHTML 1.1.")
(defun ihtml-completing-markup ()
  "Markup selection with completion."
  (interactive "*")
  (let ((tags (if *ihtml-xhtml11* *ihtml-tags-xhtml11* *ihtml-tags*))
        (tags-empty (if *ihtml-xhtml11* *ihtml-tags-empty-xhtml11* *ihtml-tags-empty*)))
    (let* ((tag (completing-read "tag: " tags :must-match t))
           (ihtml-tag (concat "ihtml-" tag)))
      (if (and (find-symbol ihtml-tag) (fboundp (intern ihtml-tag))) (funcall (intern ihtml-tag))
        (if (member tag tags-empty :test 'string=) (ihtml-markup-empty tag)
          (ihtml-markup-selection tag))))))

#|
(remove-if
 #'null
 (mapcar
  (lambda (sym)
    (if (find-symbol (concat "ihtml-" (string sym)))
        (values)
      sym))
  *ihtml-tags*))
(html head title meta style body form input button label legend select optgroup option textarea table caption thead tfoot tbody col colgroup tr th td map area)
|#



;;; (18) header elements
;;; 2007-09-12, 2014-10-28

;;; http://www.asahi-net.or.jp/~sd5a-ucd/rec-html401j/types.html#type-links
(defvar *ihtml-link-types* '(alternate stylesheet start next prev contents index glossary copyright chapter section subsection appendix help bookmark)
  "rel types of <link>.")
(ihtml-defun-markup-empty link
  (list "rel" (completing-read "rel: " *ihtml-link-types*) "href" (ihtml-string-to-uri (read-file-name "href: " :default "")))
  :doc "Insert <link rel=\"...\" href=\"...\">.") ; problem: only one link type can be inputted.
(defvar *ihtml-meta-http-equiv* '(Content-Type Content-Style-Type Content-Script-Type)
  "List of <meta http-equiv=\"...\">.") ; by "HTML/Elements/meta - W3C Wiki" http://www.w3.org/wiki/HTML/Elements/meta
(defvar *ihtml-meta-name* '(Author Copyright Date Description Keywords Modified Robots)
  "List of <meta name=\"...\">.")
(ihtml-defun-markup-empty meta
  (append (if (y-or-n-p "http-equiv?")
              (list "http-equiv" (completing-read "http-equiv: " *ihtml-meta-http-equiv*))
            (list "name" (completing-read "name: " *ihtml-meta-name*)))
          (list "content" (read-string "content: ")))
  :doc "Insert <meta http-equiv=\"...\" content=\"...\"> or <meta name=\"...\" content=\"...\">.")



#|
(defvar *ihtml-media-descriptors* '(screen tty tv projection handheld print braille aural all))
(defun ihtml-link ()
  (interactive "*")
  (let (attr
        (rel (completing-read "rel: " *ihtml-link-types*))
        (title (read-string "title: "))
        (media (completing-read "media: " *ihtml-media-descriptors*)))
    (ihtml-markup-empty "link" attr)))
(defmacro ihtml-implied-attr (attr attr-value-func)
  (let ((attr-value (gensym)))
    `(let ((,attr-value ,attr-value-func))
       (if (string/= ,attr-value "") (list ,attr ,attr-value)
         (values)))))

(pme (ihtml-implied-attr "title" (read-string "title: ")))
(let ((#:G3 (read-string "title: ")))
 (if (string/= #:G3 "")
 (list "title" #:G3)
 (values)))
nil

(list (ihtml-implied-attr "title" (read-string "title: ")) (ihtml-implied-attr "cite" (read-string "cite: ")))
(nil nil)
|#



;;; (19) ihtml-escape-selection
;;; 2007-09-14, 2014-10-28

(defun replace-regexp-whole (regexp replacement &optional case-fold)
  "replace-regexp for whole buffer."
  (interactive "*sReplace regexp: \nswith: ")
  (save-excursion
    (goto-char (point-min))
    (replace-regexp (compile-regexp regexp case-fold) replacement t)))
(dolist
    (symbol-body
     '((ihtml-escape-selection
        . ((replace-regexp-whole "&" "&amp;")
           (replace-regexp-whole "<" "&lt;")
           (replace-regexp-whole ">" "&gt;")
           (replace-regexp-whole "\"" "&quot;")))
       (ihtml-unescape-selection
        . ((replace-regexp-whole "&quot;" "\"")
           (replace-regexp-whole "&gt;" ">")
           (replace-regexp-whole "&lt;" "<")
           (replace-regexp-whole "&amp;" "&")))))
  (eval
   (let ((symbol (car symbol-body))
         (body (cdr symbol-body)))
     `(defun ,symbol ()
        ,(format nil "~A selection. e.g. & <-> &amp;" (second (split-string (string symbol) #\-)))
        (interactive "*")
        (ed::map-selection
         (lambda (start end)
           (save-excursion
             (save-restriction
               (narrow-to-region start end)
               ,@body))))))))



;;; (20) ihtml-a-fragment
;;; 2012-04-04, 2014-10-28

(defun ihtml-id-list ()
  "Return list of attribute values of id (id=\"...\")."
  (let ((regexp (compile-regexp "\\bid=\"?\\([^>\"]*\\)\"?" nil))
        fragments)
    (save-excursion
      (goto-char (point-min))
      (while (scan-buffer regexp :no-dup t)
        (push (match-string 1) fragments)))
    (nreverse fragments)))
(defun ihtml-a-fragment ()
  "Markup selection by <a href=\"#...\">."
  (interactive "*")
  (ihtml-markup-selection "a" (list "href" (ihtml-string-to-uri (concat "#" (completing-read "href: #" (ihtml-id-list) :must-match nil))))))



;;; (22) ihtml-implied-attr
;;; 2014-10-30

(defmacro ihtml-implied-attr (attr-name)
  "Return list only if parameter of attr-name is given. e.g. (ihtml-defun-markup-selection legend (let ((accesskey (read-string \"accesskey: \"))) (ihtml-implied-attr accesskey)))"
  `(when (string/= ,attr-name "") (list ,(string attr-name) ,attr-name)))



;;; (21) form elements
;;; 2014-10-29, 2014-10-30

(ihtml-defun-markup-selection form
  (list "action" (ihtml-relative-path (read-file-name "action: ")) "method" (completing-read "method: " '(get post) :must-match t)#| "enctype" (completing-read "enctype: " '(application/x-www-form-urlencoded multipart/form-data))|#)
  :doc "Insert <form action=\"\" method=\"\">.")
(dolist (tag '(fieldset))
  (eval `(ihtml-defun-markup-selection ,tag nil :doc ,(format nil "Markup selection by <~A>." tag))))
(defvar *ihtml-InputType* '(checkbox file hidden password radio reset submit text)
  "List of <input type=\"...\">.")
(ihtml-defun-markup-empty input
  (let ((type (completing-read "type: " *ihtml-InputType* :must-match t))
        (name (read-string "name: ")))
    (append
     (list "type" type)
     (ihtml-implied-attr name)
     (when (string-not-equal type "file")
       (list "value" (read-string "value: ")))
     (when (or (string-equal type "text") (string-equal type "password"))
       (let ((size (read-string "size: "))
             (maxlength (read-string "maxlength: ")))
         (append
          (ihtml-implied-attr size)
          (ihtml-implied-attr maxlength))))
     (when (or (string-equal type "checkbox") (string-equal type "radio"))
       (when (y-or-n-p "checked?") (list "checked" nil)))))
  :doc "Insert <input type=\"\" name=\"\" ...>.")
(ihtml-defun-markup-selection button
  (let ((type (completing-read "type: " '(button submit reset) :must-match t))
        (name (read-string "name: "))
        (value (read-string "value: ")))
    (append
     (list "type" type)
     (ihtml-implied-attr name)
     (ihtml-implied-attr value)))
  :doc "Markup selection by <button>.")
(ihtml-defun-markup-selection label
  (let ((for (completing-read "for: " (ihtml-id-list)))
        (accesskey (read-string "accesskey: ")))
    (append
     (ihtml-implied-attr for)
     (ihtml-implied-attr accesskey)))
  :doc "Markup selection by <label>.")
(ihtml-defun-markup-selection legend
  (let ((accesskey (read-string "accesskey: ")))
    (ihtml-implied-attr accesskey))
  :doc "Markup selection by <legend>.")
(ihtml-defun-markup-selection select
  (let ((name (read-string "name: "))
        (size (read-string "size: "))
        (multiple (y-or-n-p "multiple?")))
    (append
     (ihtml-implied-attr name)
     (ihtml-implied-attr size)
     (when multiple (list "multiple" nil))))
  :doc "Markup selection by <select>.")
(ihtml-defun-markup-selection optgroup
  (list "label" (read-string "label: "))
  :doc "Markup selection by <optgroup label=\"...\">.")
(ihtml-defun-markup-selection option
  (let ((label (read-string "label: "))
        (value (read-string "value: "))
        (selected (y-or-n-p "selected?")))
    (append
     (ihtml-implied-attr label)
     (ihtml-implied-attr value)
     (when selected (list "selected" nil))))
  :doc "Markup selection by <option>.")
(ihtml-defun-markup-selection textarea
  (let ((name (read-string "name: "))
        (rows (read-string "rows: "))
        (cols (read-string "cols: ")))
    (append
     (ihtml-implied-attr name)
     (list "rows" rows
           "cols" cols)))
  :doc "Markup selection by <textarea name=\"...\" rows=\"...\" cols=\"...\">.")

#| form elements in DTD !ATTLIST order
form
fieldset
input
button
label
legend
select
optgroup
option
textarea
|#



;;; (23) ihtml-mode
;;; 2021-09-08
(defun ihtml-mode ()
  "Mode for ISO-HTML and XHTML 1.1."
  (interactive)
  (kill-all-local-variables)
  (setq buffer-mode 'ihtml-mode)
  (setq mode-name "iHTML")
  (use-keymap *ihtml-keymap*)
  (ihtml-define-keymap))

#|
;;; in siteinit.l or .xyzzy

;;; use both html+-mode and ihtml
(in-package "editor")
(export 'ed::html+-mode "ed")
;(autoload 'html+-mode "html+-mode" t)
(autoload 'html+-mode "html+-mode-u" t)
(pushnew '("\\.s?html?$" . html+-mode) *auto-mode-alist* :test 'equal)
(in-package "user")

;;; use ihtml only
(pushnew '("\\.s?html?$" . ihtml-mode) *auto-mode-alist* :test 'equal)
|#



;;; next

#|
ihtml-escape-region
ihtml-unescape-region
form elements -> implemented
ruby
table elements
|#



#|
;;; iso-html all elements (order: dtd, element)
html
head
title
link
meta
style
;;; %Preparation;
; pre-html
; div1
; div2
; div3
; div4
; div5
; div6
;;; /%Preparation;
body
h1
h2
h3
h4
h5
h6
div
address
p
ol
ul
li
dl
dt
dd
pre
blockquote
q
fieldset
form
input
button
label
legend
select
optgroup
option
textarea
hr
table
caption
thead
tfoot
tbody
col
colgroup
tr
th
td
abbr
acronym
cite
code
dfn
em
kbd
samp
strong
var
b
i
sub
sup
tt
a
img
object
param
br
map
area
span
del
ins
bdo



;;; iso-html elements (order: dtd, entity)
;;; special


;;; logical.styles
;;; physical.styles
;;; block
;;; form.fields
;;; 
;;; 
;;; 
;;; 
|#
